/* 
 * packet-saratoga.c
 * Routines for saratoga dissection
 * Copyright 2011, Charles Smith <charlesetsmith@me.com>
 *
 * $Id: packet-saratoga.c,v 1.4 2011/08/05 05:36:58 charlesetsmith Exp $
 *
 * Wireshark - Network traffic analyzer
 * By Gerald Combs <gerald@wireshark.org>
 * Copyright 1998 Gerald Combs
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

/*
 * Saratoga Dissector for Wireshark
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <glib.h>
#include <stdlib.h>
#include <ctype.h>
#include <epan/packet.h>
#include <epan/ipproto.h>
#include <epan/prefs.h>

#define SAR_PORT 7542

/*
 * The number of seconds betwen 1/1/1970 & 1/1/2000 including leap seconds
 * So we can convert between posix ctime and bundle protocol time as used
 * in directory entries
 */
#define BUNDLE_CONVERSION 946684800

static int	proto_saratoga = -1;

/*
 * Flags in Saratoga 32 bit header field
 */
static int	hf_sar_flags = -1;
static int	hf_sar_version = -1;
static int	hf_sar_packettype = -1;
static int	hf_sar_descriptor = -1;
static int	hf_sar_bundle = -1;
static int	hf_sar_stream = -1;
static int	hf_sar_txwill = -1;
static int	hf_sar_rxwill = -1;
static int	hf_sar_udp = -1;
static int	hf_sar_fsp = -1;
static int	hf_sar_fspd = -1;
static int	hf_sar_tran = -1;
static int	hf_sar_tfr = -1;
static int	hf_sar_prog = -1;
static int	hf_sar_rel = -1;
static int	hf_sar_csumlen = -1;
static int	hf_sar_csum = -1;
static int	hf_sar_reqstatus = -1;
static int	hf_sar_tstamp = -1;
static int	hf_sar_eod = -1;
static int	hf_sar_met = -1;
static int	hf_sar_allholes = -1;
static int	hf_sar_requestby = -1;
static int	hf_sar_beacon_eid = -1;
static int	hf_sar_filepath = -1;
static int	hf_sar_statusflag = -1;

/*
 * Directory Entry Flags
 */
static int	hf_sar_dirstart = -1;
static int	hf_sar_dirprop = -1;
static int	hf_sar_dirdesc = -1;
static int	hf_sar_dirforb = -1;
static int	hf_sar_dirrel = -1;

static gint	ett_saratoga = -1;
static gint	ett_sar_beacon = -1;
static gint	ett_sar_request = -1;
static gint	ett_sar_metadata = -1;
static gint	ett_sar_data = -1;
static gint	ett_sar_status = -1;

/*
 * Version is the high order 2 bits of the first byte
 * BEACON, REQUEST, METADATA, DATA, STATUS
 */
#define SAR_VERSION_SHIFT 29
#define SAR_VERSION_MASK (0x07 << SAR_VERSION_SHIFT)

static const value_string sar_version_names[] =
{
	{ 0, "V0"},
	{ 1, "V1"},
	{ 2, "V2 Invalid"},
	{ 3, "V3 Invalid"},
	{ 4, "V4 Invalid"},
	{ 5, "V5 Invalid"},
	{ 6, "V6 Invalid"},
	{ 7, "V7 Invalid"},
	{ 0, NULL }
};

/*
 * Packettype is the low order 6 bits of the first byte
 * BEACON, REQUEST, METADATA, DATA, STATUS
 */
#define SAR_PACKETTYPE_SHIFT 24 
#define SAR_PACKETTYPE_MASK (0x1F << SAR_PACKETTYPE_SHIFT)
#define SAR_FRAME_TYPE(flag)	(flag & SAR_PACKETTYPE_MASK) >> SAR_PACKETTYPE_SHIFT

#define SAR_BEACON	0
#define SAR_REQUEST	1
#define SAR_METADATA	2
#define SAR_DATA	3
#define SAR_STATUS	4

static const value_string sar_packettype_names[] =
{
	{ SAR_BEACON, "Beacon" },
	{ SAR_REQUEST, "Request" },
	{ SAR_METADATA, "Metadata" },
	{ SAR_DATA, "Data" },
	{ SAR_STATUS, "Status" },
	{ 0, NULL }
};

/*
 * Descriptor is the high order 2 bits of the second byte
 * BEACON, REQUEST, METADATA, DATA, STATUS
 */
#define SAR_DESC_MASK 0xC00000
#define SAR_DESC_SHIFT 22 

#define SAR_DESC_TYPE(flag)	(flag & SAR_DESC_MASK) >> SAR_DESC_SHIFT
#define SAR_DESC_16 0
#define SAR_DESC_32 1
#define SAR_DESC_64 2
#define SAR_DESC_128 3

static const value_string sar_descriptor_names[] =
{
	{ SAR_DESC_16, "16 Bit" },
	{ SAR_DESC_32, "32 Bit" },
	{ SAR_DESC_64, "64 Bit" },
	{ SAR_DESC_128, "128 Bit" },
	{ 0, NULL }
};

/*
 * BEACON, REQUEST
 */
#define SAR_BUNDLE_SHIFT 21
#define SAR_BUNDLE_MASK (0x01 << SAR_BUNDLE_SHIFT)

static const value_string sar_bundle_names[] =
{
	{ 0, "Not Supported" },
	{ 1, "Supported" },
	{ 0, NULL }
};

/*
 * BEACON, REQUEST
 */
#define SAR_STREAM_SHIFT 20
#define SAR_STREAM_MASK (0x01 << SAR_STREAM_SHIFT)

static const value_string sar_stream_names[] =
{
	{ 0, "Not Supported" },
	{ 1, "Supported" },
	{ 0, NULL }
};

/*
 * BEACON
 */
#define SAR_TXWILL_SHIFT 18
#define SAR_TXWILL_MASK (0x03 << SAR_TXWILL_SHIFT)

static const value_string sar_txwill_names[] =
{
	{ 0, "Unwilling" },
	{ 1, "Invalid Flag" },
	{ 2, "Capable but not right now" },
	{ 3, "Willing" },
	{ 0, NULL }
};

/*
 * BEACON
 */
#define SAR_RXWILL_SHIFT 16
#define SAR_RXWILL_MASK (0x03 << SAR_RXWILL_SHIFT)

static const value_string sar_rxwill_names[] =
{
	{ 0, "Unwilling" },
	{ 1, "Invalid Flag" },
	{ 2, "Capable but not right now" },
	{ 3, "Willing" },
	{ 0, NULL }
};

/*
 * BEACON,REQUEST
 */
#define SAR_UDP_SHIFT 15
#define SAR_UDP_MASK (0x01 << SAR_UDP_SHIFT)

static const value_string sar_udp_names[] =
{
	{ 0, "UDP Only" },
	{ 1, "UDP or UDP Lite" },
	{ 0, NULL }
};

/*
 * BEACON
 */
#define SAR_FSP_SHIFT 14
#define SAR_FSP_MASK (0x01 << SAR_FSP_SHIFT)
#define SAR_FSP_TYPE(flag)	(flag & SAR_FSP_MASK) >> SAR_FSP_SHIFT
#define SAR_FSP_NOTADVERT	0
#define SAR_FSP_ADVERT		1

static const value_string sar_fsp_names[] =
{
	{ SAR_FSP_NOTADVERT, "Not Advertised" },
	{ SAR_FSP_ADVERT, "Advertised" },
	{ 0, NULL }
};

/*
 * BEACON
 */
#define SAR_FSPD_SHIFT 12
#define SAR_FSPD_MASK (0x03 << SAR_FSPD_SHIFT)
#define SAR_FSPD_TYPE(flag)	(flag & SAR_FSPD_MASK) >> SAR_FSPD_SHIFT

static const value_string sar_fspd_names[] =
{
	{ SAR_DESC_16, "16 bit" },
	{ SAR_DESC_32, "32 bit" },
	{ SAR_DESC_64, "64 bit" },
	{ SAR_DESC_128, "128 bit" },
	{ 0, NULL }
};

#define SAR_TRAN_SHIFT 0
#define SAR_TRAN_MASK (0xFF << SAR_TRAN_SHIFT)

/*
 * REQUEST
 */
static const value_string sar_tran_names[] =
{
	{ 0, "No Action" },
	{ 1, "Get File" },
	{ 2, "Put File" },
	{ 3, "Get and Delete File" },
	{ 4, "Put and Delete File" },
	{ 5, "Delete" },
	{ 6, "Get Directory" },
	{ 0, NULL }
};

#define SAR_TFR_SHIFT 20
#define SAR_TFR_MASK (0x03 << SAR_TFR_SHIFT)
#define SAR_TFR_TYPE(flag)	(flag & SAR_TFR_MASK) >> SAR_TFR_SHIFT
#define SAR_TFR_FILE	0
#define SAR_TFR_DIR	1
#define SAR_TFR_BUNDLE	2
#define SAR_TFR_STREAM	3

/*
 * METADATA, DATA
 */
static const value_string sar_tfr_names[] =
{
	{ SAR_TFR_FILE, "File" },
	{ SAR_TFR_DIR, "Directory" },
	{ SAR_TFR_BUNDLE, "Bundle" },
	{ SAR_TFR_STREAM, "Stream" },
	{ 0, NULL }
};

/*
 * METADATA
 */
#define SAR_PROG_SHIFT 19
#define SAR_PROG_MASK (0x01 << SAR_PROG_SHIFT)

static const value_string sar_prog_names[] =
{
	{ 0, "In Progress" },
	{ 1, "Terminated" },
	{ 0, NULL },
};

/*
 * METADATA
 */
#define SAR_REL_SHIFT 18
#define SAR_REL_MASK (0x01 << SAR_REL_SHIFT)

static const value_string sar_rel_names[] =
{
	{ 0, "With no errors via UDP" },
	{ 1, "Errors tolerated" },
	{ 0, NULL },
};

/*
 * METADATA
 * Multiple of 32 bits
 */
#define SAR_CSUMLEN_SHIFT 4
#define SAR_CSUMLEN_MASK (0x0F << SAR_CSUMLEN_SHIFT)
#define SAR_CSUMLEN_TYPE(flag)	(flag & SAR_CSUMLEN_MASK) >> SAR_CSUMLEN_SHIFT

static const value_string sar_csumlen_names[] =
{
	{ 0, "0" },
	{ 1, "32 bits" },
	{ 2, "64 bits" },
	{ 3, "96 bits" },
	{ 4, "128 bits" },
	{ 5, "160 bits" },
	{ 6, "192 bits" },
	{ 7, "224 bits" },
	{ 8, "256 bits" },
	{ 9, "288 bits" },
	{ 10, "320 bits" },
	{ 11, "352 bits" },
	{ 12, "384 bits" },
	{ 13, "416 bits" },
	{ 14, "448 bits" },
	{ 15, "480 bits" }
};

/*
 * METADATA
 */
#define SAR_CSUM_SHIFT 0
#define SAR_CSUM_MASK (0x0F << SAR_CSUM_SHIFT)
#define SAR_CSUM_TYPE(flag)	(flag & SAR_CSUM_MASK) >> SAR_CSUM_SHIFT
#define SAR_CSUM_NONE		0
#define SAR_CSUM_CRC32		1
#define SAR_CSUM_MD5		2
#define SAR_CSUM_SHA1		3


static const value_string sar_csum_names[] =
{
	{ 0, "None" },
	{ 1, "32-bit CRC32" },
	{ 2, "128-bit MD5" },
	{ 3, "160-bit SHA-1" },
	{ 0, NULL }
};

/*
 * DATA, STATUS
 */
#define SAR_TSTAMP_SHIFT 19
#define SAR_TSTAMP_MASK (0x01 << SAR_TSTAMP_SHIFT)
#define SAR_TSTAMP_TYPE(flag)	(flag & SAR_TSTAMP_MASK) >> SAR_TSTAMP_SHIFT
#define SAR_TSTAMP_DISABLED	0
#define SAR_TSTAMP_ENABLED	1

static const value_string sar_tstamp_names[] =
{
	{ SAR_TSTAMP_DISABLED, "Disabled" },
	{ SAR_TSTAMP_ENABLED, "Enabled" },
	{ 0, NULL }
};

/*
 * DATA
 */
#define SAR_REQSTATUS_SHIFT 16
#define SAR_REQSTATUS_MASK (0x01 << SAR_REQSTATUS_SHIFT)

static const value_string sar_reqstatus_names[] =
{
	{ 0, "Not Requested" },
	{ 1, "Requested" },
	{ 0, NULL }
};

/*
 * DATA
 */
#define SAR_EOD_SHIFT 15
#define SAR_EOD_MASK (0x01 << SAR_EOD_SHIFT)

static const value_string sar_eod_names[] =
{
	{ 0, "No" },
	{ 1, "Yes" },
	{ 0, NULL }
};

/*
 * STATUS
 */
#define SAR_MET_SHIFT 18
#define SAR_MET_MASK (0x01 << SAR_MET_SHIFT)

static const value_string sar_met_names[] =
{
	{ 0, "Received" },
	{ 1, "Not Received" },
	{ 0, NULL }
};

/*
 * STATUS
 */
#define SAR_ALLHOLES_SHIFT 17
#define SAR_ALLHOLES_MASK (0x01 << SAR_ALLHOLES_SHIFT)

static const value_string sar_allholes_names[] =
{
	{ 0, "Complete List in Frame" },
	{ 1, "Incomplete List" },
	{ 0, NULL }
};

/*
 * STATUS
 */
#define SAR_REQUESTBY_SHIFT 16
#define SAR_REQUESTBY_MASK (0x01 << SAR_REQUESTBY_SHIFT)

static const value_string sar_requestby_names[] =
{
	{ 0, "Requested by sender" },
	{ 1, "Sent Voluntarily" },
	{ 0, NULL }
};

/*
 * Handle Saratoga Timestamps
 * 16 Byte Field - Low order 3 bits define type
 */
#define SAR_TIMESTAMP_MASK	0x07

#define SAR_TIMESTAMP_LOCAL	0
#define SAR_TIMESTAMP_32	1
#define SAR_TIMESTAMP_64	2
#define SAR_TIMESTAMP_32_32	3
#define SAR_TIMESTAMP_64_32	4

/*
 * Handle Saratoga Directory Entry Properties Flags
 * 16 Bits Long
 */

/*
 * First bit is always set to 1 to delimit record
 */
#define DIR_START_SHIFT 15
#define DIR_START_MASK (0x01 << DIR_START_SHIFT)
#define DIR_START_TYPE(flag)	(flag & DIR_START_MASK) >> DIR_START_SHIFT
static const value_string sar_dirstart_names[] =
{
	{ 0, "Invalid should always be 1" },
	{ 1, "Start of Directory Record" },
	{ 0, NULL }
};

#define DIR_PROP_SHIFT	8
#define DIR_PROP_MASK	(0x03 << DIR_PROP_SHIFT)
#define DIR_PROP_TYPE(flag)	(flag & DIR_PROP_MASK) >> DIR_PROP_SHIFT

static const value_string sar_dirprop_names[] =
{
	{ 0, "Normal File" },
	{ 1, "Normal Directory" },
	{ 2, "Special File" },
	{ 3, "Special Directory" },
	{ 0, NULL }
};

/*
 * Directory Entry File Size Descriptor Size
 */
#define DIR_DESC_SHIFT	6
#define DIR_DESC_MASK	(0x03 << DIR_DESC_SHIFT)
#define DIR_DESC_TYPE(flag)	(flag & DIR_DESC_MASK) >> DIR_DESC_SHIFT

#define DIR_DESC_16	0
#define DIR_DESC_32	1
#define DIR_DESC_64	2
#define DIR_DESC_128	3

static const value_string sar_dirdesc_names[] =
{
	{ DIR_DESC_16, "16 Bit" },
	{ DIR_DESC_32, "32 Bit" },
	{ DIR_DESC_64, "64 Bit" },
	{ DIR_DESC_128, "128 Bit" },
	{ 0, NULL }
};

/*
 * Directory Entry are we a file or a bundle
 */
#define DIR_FORB_SHIFT	5
#define DIR_FORB_MASK (0x01 << DIR_FORB_SHIFT)
#define DIR_FORB_TYPE(flag)	(flag & DIR_FORB_MASK) >> DIR_FORB_SHIFT

static const value_string sar_dirforb_names[] =
{
	{ 0, "Is a File" },
	{ 1, "Is a Bundle" },
	{ 0, NULL }
};

/*
 * Directory Entry How can we be delivered
 */
#define DIR_REL_SHIFT	2
#define DIR_REL_MASK (0x01 << DIR_REL_SHIFT)
#define DIR_REL_TYPE(flag)	(flag & DIR_REL_MASK) >> DIR_REL_SHIFT

static const value_string sar_dirrel_names[] =
{
	{ 0, "Reliably via UDP Only" },
	{ 1, "Unreliably or Partly Reliable with UDP or UDP Lite" },
	{ 0, NULL }
};

/*
 * Status Flags
 */
#define SAR_STATUSFLAG_SHIFT	0
#define SAR_STATUSFLAG_MASK	0xFF
#define SAR_STATUSFLAG_TYPE(flag)	(flag & SAR_STATUSFLAG_MASK)

static const value_string sar_statusflag_names[] =
{
	{ 0, "Success, No Errors" },
	{ 1, "Unspecified Error" },
	{ 2, "Unable to send file due to resource constraints" },
	{ 3, "Unable to receive file due to resource constraints" },
	{ 4, "File not found" },
	{ 5, "Access Denied" },
	{ 6, "Unknown ID field for transaction" },
	{ 7, "Did not delete file" },
	{ 8, "File length longer than REQUEST indicates support for" },
	{ 9, "File offset descriptors do not match expected use or file length" },
	{ 0x0A, "Unsupported packet type received" },
	{ 0x0B, "DATA flag bits describing transfer have changed unexpectedly" },
	{ 0x0C, "Receiver is no longer interested in receiving this file" },
	{ 0x0D, "Receiver wants sender to pause its transfer" },
	{ 0x0E, "Receiver wants sender to resume a previously-paused transfer" },
	{ 0x0F, "File is in use" },
	{ 0x10, "METADATA required before transfer can be accepted" },
	{ 0x11, "Unsupported REQUEST type" },
	{ 0x12, "REQUEST terminated due to an internal timeout" },
	{ 0x13, "STATUS error message received unexpectedly, REQUEST terminated" },
	{ 0, NULL }
};

/*
 * Convert the saratoga (bundle) protocol time to a posix time for display
 * using the ctime routines
 */
guint32
bundletoposix(guint32 bundletime)
{
	guint32        posixtime;

	/*
	 * Of course this is not portable once we wrap around
	 * with the 32 bit time
	 */
	posixtime = bundletime + BUNDLE_CONVERSION;
	return(posixtime);
}

/*
 * Saratoga is all about descriptor sizes so lets handle their outputs uniformly
 * 16 bit descritpror
 */
void
proto_tree_add_desc16(proto_tree *tree, tvbuff_t *tvb, gint offset, 
	char *title, guint16 descriptor)
{
	proto_tree_add_text(tree, tvb, offset, 2, "%s 16 bit: %" G_GINT16_MODIFIER "u",
		title,
		descriptor);
}

/* 
 * 32 bit descritpror
 */
void
proto_tree_add_desc32(proto_tree *tree, tvbuff_t *tvb, gint offset, 
	char *title, guint32 descriptor)
{
	proto_tree_add_text(tree, tvb, offset, 4, "%s 32 bit: %" G_GINT32_MODIFIER "u",
		title,
		descriptor);
}

/* 
 * 64 bit descritpror
 */
void
proto_tree_add_desc64(proto_tree *tree, tvbuff_t *tvb, gint offset, 
	char *title, guint64 descriptor)
{
	proto_tree_add_text(tree, tvb, offset, 8, "%s 64 bit: %" G_GINT64_MODIFIER "u",
		title,
		descriptor);
}

/* 
 * 128 bit descriptor - Not yet
 */
void
proto_tree_add_desc128(proto_tree *tree, tvbuff_t *tvb, gint offset,
	char *title)
{
	proto_tree_add_text(tree, tvb, offset, 16, "%s 128 bit: Not supported",
		title);
}

/* 
 * 16 bit hole
 */
void
proto_tree_add_hole16(proto_tree *tree, tvbuff_t *tvb, gint offset,
	gint holenumb, guint16 start, guint16 end)
{
	char	fmt[128];

	char *fmt1 = "Hole %d 16 bit: From %" G_GINT16_MODIFIER "u";
	char *fmt2 = " To %" G_GINT16_MODIFIER "u";
	char *fmt3 = " Length %" G_GINT16_MODIFIER "u";
	
	strcpy(fmt, fmt1);
	strcat(fmt, fmt2);
	strcat(fmt, fmt3);
	
	proto_tree_add_text(tree, tvb, offset, 4, fmt,
		holenumb,
		start,
		end, 
		end - start);
}

/* 
 * 32 bit hole
 */
void
proto_tree_add_hole32(proto_tree *tree, tvbuff_t *tvb, gint offset,
	gint holenumb, guint32 start, guint32 end)
{
	char	fmt[128];

	char *fmt1 = "Hole %d 32 bit: From %" G_GINT32_MODIFIER "u";
	char *fmt2 = " To %" G_GINT32_MODIFIER "u";
	char *fmt3 = " Length %" G_GINT32_MODIFIER "u";
	
	strcpy(fmt, fmt1);
	strcat(fmt, fmt2);
	strcat(fmt, fmt3);
	
	proto_tree_add_text(tree, tvb, offset, 8, fmt,
		holenumb,
		start,
		end, 
		end - start);
}

/* 
 * 64 bit hole
 */
void
proto_tree_add_hole64(proto_tree *tree, tvbuff_t *tvb, gint offset,
	gint holenumb, guint64 start, guint64 end)
{
	char	fmt[128];

	char *fmt1 = "Hole %d 32 bit: From %" G_GINT64_MODIFIER "u";
	char *fmt2 = " To %" G_GINT64_MODIFIER "u";
	char *fmt3 = " Length %" G_GINT64_MODIFIER "u";
	
	strcpy(fmt, fmt1);
	strcat(fmt, fmt2);
	strcat(fmt, fmt3);
	
	proto_tree_add_text(tree, tvb, offset, 16, fmt,
		holenumb,
		start,
		end, 
		end - start);
}

/* 
 * 128 bit hole - Not yet
 */
void
proto_tree_add_hole128(proto_tree *tree, tvbuff_t *tvb, gint offset,
	gint holenumb)
{
	proto_tree_add_text(tree, tvb, offset, 8, "Hole %d 128 bit: Not supported",
		holenumb);
}

/*
 * Transaction ID 32 bit Unsigned Integer
 */
void
proto_tree_add_tranid(proto_tree *tree, tvbuff_t *tvb, gint offset, 
	char *title, guint32 descriptor)
{
	proto_tree_add_text(tree, tvb, offset, 4, "%s: %08X" G_GINT32_MODIFIER ,
		title,
		descriptor);
}


void
proto_tree_add_ctime(proto_tree *tree, tvbuff_t *tvb, gint offset,
	char *title, guint32 thetime)
{
	static char	s[26];
	guint32 ct = thetime;

	/*
	 * Convert it to posix time for display
	 */
	ct = bundletoposix(ct);	
	ctime_r((time_t *) &ct, s);
	s[strlen(s) -1] = '\0';
	proto_tree_add_text(tree, tvb, offset, 4, "%s: %s",
		title,
		s);
}

/*
 * Dissect a Checksum
 */

void
proto_tree_add_csum(proto_tree *tree, tvbuff_t *tvb, gint offset, 
	char *title, gint len32)
{
	gint i;
	guint32 csum[16]; /* Max 16 * 32 bit Checksums */
	char	str[512];
	char	tmp[16];

	strcpy(str, title);
	for (i = 0; i < len32; i++)
	{
		csum[i] = tvb_get_ntohl(tvb, offset);
		offset += 4;
		g_snprintf(tmp, 16, "%08x-", csum[i]);
		strcat(str, tmp);
	}
	str[strlen(str) - 1] = '\0';
	proto_tree_add_text(tree, tvb, 
		offset - len32 * 4, len32 * 4, "%s", str);
}

/*
 * If we have all printable characters in the buffer then print them
 * otherwise send them out as hex digits
 */
void
proto_tree_add_printtext(proto_tree *tree, tvbuff_t *tvb, gint offset,
	char *fmt, gint blen)
{
	gint	i = offset;
	guint8 *id = NULL;
	guint8	c;

	char	sfmt[128];

	strcpy(sfmt, fmt);
	strcat(sfmt, " %s\n");
	if (blen <= 0)
		return;
	for (i = 0; i < blen - 1; i++)
	{
		c = tvb_get_guint8(tvb, offset +i);
		if (!isprint((char) c))
		{
			/*
			 * We have a non printable character so output a
			 * hex string instead
			 */
			id = tvb_bytes_to_str_punct(tvb, 
				offset,
				blen,
				'-');
			proto_tree_add_text(tree, tvb, offset,
				blen,
				sfmt,
				id);
			return;
		}
	}
	/*
	 * We have a fully printable ascii string
	 */
	id = tvb_get_string(tvb, offset, blen);
	proto_tree_add_text(tree, tvb, offset,
		blen,
		sfmt,
		id);
}

/*
 * Dissect a Directory Entry
 * Return how may bytes it is so we can adjust the offset
 */
gint
proto_tree_add_dir(proto_tree *tree, tvbuff_t *tvb, gint offset)
{
	guint16 dirflags = 0;
	guint16 tmp_16 = 0;
	guint32 tmp_32 = 0;
	guint64 tmp_64 = 0;
	guint8 *filepath = NULL;
	gint blen = 0;
	gint origoffset = offset;
	/*
	 * Get Directory Entry Properties Flags
	 */
	dirflags = tvb_get_ntohs(tvb, offset);
	proto_tree_add_item(tree, hf_sar_dirstart,
			tvb, offset, 2, ENC_BIG_ENDIAN);
	proto_tree_add_item(tree, hf_sar_dirprop,
			tvb, offset, 2, ENC_BIG_ENDIAN);
	proto_tree_add_item(tree, hf_sar_dirdesc,
			tvb, offset, 2, ENC_BIG_ENDIAN);
	proto_tree_add_item(tree, hf_sar_dirforb,
			tvb, offset, 2, ENC_BIG_ENDIAN);
	proto_tree_add_item(tree, hf_sar_dirrel,
			tvb, offset, 2, ENC_BIG_ENDIAN);
	offset += 2;

	/*
	 * Get Directory Entry Filesize Info
	 */
	switch(DIR_DESC_TYPE(dirflags))
	{
	case DIR_DESC_16:
		tmp_16 = tvb_get_ntohs(tvb, offset);
		proto_tree_add_desc16(tree,
			tvb, offset, "File Size (bytes)",
			tmp_16);
		offset += 2;
		break;
	case DIR_DESC_32:
		tmp_32 = tvb_get_ntohl(tvb, offset);
		proto_tree_add_desc32(tree,
			tvb, offset, "File Size (bytes)",
			tmp_32);
		offset += 4;
		break;
	case DIR_DESC_64:
		tmp_64 = tvb_get_ntoh64(tvb, offset);
		proto_tree_add_desc64(tree,
			tvb, offset, "File Size (bytes)",
			tmp_64);
		offset += 8;
		break;
	case DIR_DESC_128:
		/* No 128 bit support yet!!!! */
		proto_tree_add_desc128(tree,
			tvb, offset, "File Size (bytes)");
		offset += 16;
		break;
	}
	/*
	 * Get Directory Entry Mtime
	 */
	tmp_32 = tvb_get_ntohl(tvb, offset);
	proto_tree_add_ctime(tree,
		tvb, offset, "Mtime",
		tmp_32);
	offset += 4;
	/*
	 * Get Directory Entry Ctime
	 */
	tmp_32 = tvb_get_ntohl(tvb, offset);
	proto_tree_add_ctime(tree,
		tvb, offset, "Ctime",
		tmp_32);
	offset += 4;
	/*
	 * Get Directory Entry File / Path Name
	 */
	filepath = tvb_get_stringz(tvb, offset, &blen);
	proto_tree_add_string(tree, 
		hf_sar_filepath, 
		tvb, offset, blen,
		filepath);
	g_free(filepath);
	offset += blen;
	/* 
	 * Skip along null padding to next directory record
	 */
	
	while (tvb_length_remaining(tvb, offset) > 0 &&
		tvb_get_guint8(tvb, offset) == 0)
			offset++;
	return(offset - origoffset);
}

void
proto_register_saratoga(void)
{
	static hf_register_info hf[] =
	{
		{
			&hf_sar_flags,
			{
				"Flags", "saratoga.flags",
				FT_UINT32, BASE_HEX,
				NULL, 0,
				NULL, HFILL
			}
		},
		{
			&hf_sar_version,
			{
				"Version", "saratoga.version",
				FT_UINT32, BASE_HEX,
				VALS(sar_version_names), SAR_VERSION_MASK,
				NULL, HFILL
			}
		},
		{
			&hf_sar_packettype,
			{
				"Packet Type", "saratoga.packettype",
				FT_UINT32, BASE_HEX,
				VALS(sar_packettype_names), SAR_PACKETTYPE_MASK,
				NULL, HFILL
			}
		},
		{
			&hf_sar_descriptor,
			{
				"Descriptor Size", "saratoga.descriptor",
				FT_UINT32, BASE_HEX,
				VALS(sar_descriptor_names), SAR_DESC_MASK,
				NULL, HFILL
			}
		},
		{
			&hf_sar_bundle,
			{
				"Bundles", "saratoga.bundle",
				FT_UINT32, BASE_HEX,
				VALS(sar_bundle_names), SAR_BUNDLE_MASK,
				NULL, HFILL
			}
		},
		{
			&hf_sar_stream,
			{
				"Streams", "saratoga.stream",
				FT_UINT32, BASE_HEX,
				VALS(sar_stream_names), SAR_STREAM_MASK,
				NULL, HFILL
			}
		},
		{
			&hf_sar_txwill,
			{
				"Transmit", "saratoga.txwill",
				FT_UINT32, BASE_HEX,
				VALS(sar_txwill_names), SAR_TXWILL_MASK,
				NULL, HFILL
			}
		},
		{
			&hf_sar_rxwill,
			{
				"Receive", "saratoga.rxwill",
				FT_UINT32, BASE_HEX,
				VALS(sar_rxwill_names), SAR_RXWILL_MASK,
				NULL, HFILL
			}
		},
		{
			&hf_sar_udp,
			{
				"UDP Lite", "saratoga.udp",
				FT_UINT32, BASE_HEX,
				VALS(sar_udp_names), SAR_UDP_MASK,
				NULL, HFILL
			}
		},
		{
			&hf_sar_fsp,
			{
				"Free Space", "saratoga.fsp",
				FT_UINT32, BASE_HEX,
				VALS(sar_fsp_names), SAR_FSP_MASK,
				NULL, HFILL
			}
		},
		{
			&hf_sar_fspd,
			{
				"Free Space Descriptor", "saratoga.fspd",
				FT_UINT32, BASE_HEX,
				VALS(sar_fspd_names), SAR_FSPD_MASK,
				NULL, HFILL
			}
		},
		{
			&hf_sar_tran,
			{
				"Transaction Type", "saratoga.tran",
				FT_UINT32, BASE_HEX,
				VALS(sar_tran_names), SAR_TRAN_MASK,
				NULL, HFILL
			}
		},
		{
			&hf_sar_tfr,
			{
				"Transfer", "saratoga.tfr",
				FT_UINT32, BASE_HEX,
				VALS(sar_tfr_names), SAR_TFR_MASK,
				NULL, HFILL
			}
		},
		{
			&hf_sar_prog,
			{
				"Progress", "saratoga.prog",
				FT_UINT32, BASE_HEX,
				VALS(sar_prog_names), SAR_PROG_MASK,
				NULL, HFILL
			}
		},
		{
			&hf_sar_rel,
			{
				"Reliability", "saratoga.rel",
				FT_UINT32, BASE_HEX,
				VALS(sar_rel_names), SAR_REL_MASK,
				NULL, HFILL
			}
		},
		{
			&hf_sar_csumlen,
			{
				"Checksum Length", "saratoga.csum",
				FT_UINT32, BASE_HEX,
				VALS(sar_csumlen_names), SAR_CSUMLEN_MASK,
				NULL, HFILL
			}
		},
		{
			&hf_sar_csum,
			{
				"Checksum", "saratoga.csum",
				FT_UINT32, BASE_HEX,
				VALS(sar_csum_names), SAR_CSUM_MASK,
				NULL, HFILL
			}
		},
		{
			&hf_sar_tstamp,
			{
				"Timestamps", "saratoga.tstamp",
				FT_UINT32, BASE_HEX,
				VALS(sar_tstamp_names), SAR_TSTAMP_MASK,
				NULL, HFILL
			}
		},
		{
			&hf_sar_reqstatus,
			{
				"Request Status", "saratoga.reqstatus",
				FT_UINT32, BASE_HEX,
				VALS(sar_reqstatus_names), SAR_REQSTATUS_MASK,
				NULL, HFILL
			}
		},
		{
			&hf_sar_eod,
			{
				"Last Data Frame", "saratoga.eod",
				FT_UINT32, BASE_HEX,
				VALS(sar_eod_names), SAR_EOD_MASK,
				NULL, HFILL
			}
		},
		{
			&hf_sar_met,
			{
				"Metadata", "saratoga.met",
				FT_UINT32, BASE_HEX,
				VALS(sar_met_names), SAR_MET_MASK,
				NULL, HFILL
			}
		},
		{
			&hf_sar_allholes,
			{
				"Holes", "saratoga.allholes",
				FT_UINT32, BASE_HEX,
				VALS(sar_allholes_names), SAR_ALLHOLES_MASK,
				NULL, HFILL
			}
		},
		{
			&hf_sar_requestby,
			{
				"Status", "saratoga.requestby",
				FT_UINT32, BASE_HEX,
				VALS(sar_requestby_names), SAR_REQUESTBY_MASK,
				NULL, HFILL
			}
		},
		{
			&hf_sar_beacon_eid,
			{
				"EID", "saratoga.beaconeid",
				FT_STRING, BASE_NONE,
				NULL, 0,
				NULL, HFILL
			}
		},
		{
			&hf_sar_filepath,
			{
				"File or Path Name", "saratoga.filepath",
				FT_STRING, BASE_NONE,
				NULL, 0,
				NULL, HFILL
			}
		},
		{
			&hf_sar_dirstart,
			{
				"Directory Record Start", "saratoga.dirstart",
				FT_UINT16, BASE_HEX,
				VALS(sar_dirstart_names), DIR_START_MASK,
				NULL, HFILL
			}
		},
		{
			&hf_sar_dirprop,
			{
				"Directory Properties", "saratoga.dirprop",
				FT_UINT16, BASE_HEX,
				VALS(sar_dirprop_names), DIR_PROP_MASK,
				NULL, HFILL
			}
		},
		{
			&hf_sar_dirdesc,
			{
				"Directory Descriptor", "saratoga.dirdesc",
				FT_UINT16, BASE_HEX,
				VALS(sar_dirdesc_names), DIR_DESC_MASK,
				NULL, HFILL
			}
		},
		{
			&hf_sar_dirforb,
			{
				"File or Bundle", "saratoga.dirforb",
				FT_UINT16, BASE_HEX,
				VALS(sar_dirforb_names), DIR_FORB_MASK,
				NULL, HFILL
			}
		},
		{
			&hf_sar_dirrel,
			{
				"Can Be Transported", "saratoga.dirrel",
				FT_UINT16, BASE_HEX,
				VALS(sar_dirrel_names), DIR_REL_MASK,
				NULL, HFILL
			}
		},
		{
			&hf_sar_statusflag,
			{
				"Status Flag", "saratoga.statusflag",
				FT_UINT32, BASE_HEX,
				VALS(sar_statusflag_names), SAR_STATUSFLAG_MASK,
				NULL, HFILL
			}
		}
	};

	static gint *ett[] =
	{
		&ett_saratoga,
		&ett_sar_beacon,
		&ett_sar_request,
		&ett_sar_metadata,
		&ett_sar_data,
		&ett_sar_status
	};

	proto_saratoga = proto_register_protocol(
		"Saratoga Protocol", /* Name */
		"SARATOGA", /* Short Name */
		"saratoga" /* abbrev */ );

	proto_register_field_array(proto_saratoga, hf, array_length(hf));
	proto_register_subtree_array(ett, array_length(ett));
}

static void
dissect_beacon(guint32 flags, proto_tree *tree, tvbuff_t *tvb, gint offset)
{
	proto_tree *b_tree = NULL;
	gint blen = 0;
	guint16 beacon_freespace_16 = 0;
	guint32 beacon_freespace_32 = 0;
	guint64 beacon_freespace_64 = 0;
			
	if (tree == NULL)
		return;	
	b_tree = proto_item_add_subtree(tree, ett_sar_beacon);
	/*
	 * Optional Free Space Advertisement
	 */
	if (SAR_FSP_TYPE(flags) == SAR_FSP_ADVERT)
	{
		switch(SAR_FSPD_TYPE(flags))
		{
		case SAR_DESC_16:
			beacon_freespace_16 = tvb_get_ntohs(tvb, offset);
			proto_tree_add_desc16(b_tree,
				tvb, offset, "Free Space (kB)",
				beacon_freespace_16);
			offset += 2;
			break;
		case SAR_DESC_32:
			beacon_freespace_32 = tvb_get_ntohl(tvb, offset);
			proto_tree_add_desc32(b_tree,
				tvb, offset, "Free Space (kB)",
				beacon_freespace_32);
			offset += 4;
			break;
		case SAR_DESC_64:
			beacon_freespace_64 = tvb_get_ntoh64(tvb, offset);
			proto_tree_add_desc64(b_tree,
				tvb, offset, "Free Space (kB)",
				beacon_freespace_64);
			offset += 8;
			break;
		case SAR_DESC_128:
			/* No 128 bit support yet!!!! */
			proto_tree_add_desc128(b_tree,
				tvb, offset, "Free Space (kB)");
			offset += 16;
			break;
		}
	}
	/* Rest of BEACON frame is the EID 
	 * It can be any string of bytes, not UTF-8 
	 */
	blen = tvb_length_remaining(tvb, offset);
	proto_tree_add_printtext(b_tree, tvb, offset, 
		"EID:", blen);
	return;
}

static void
dissect_request(guint32 flags, proto_tree *tree, tvbuff_t *tvb, gint offset)
{
	proto_tree *r_tree = NULL;
	guint32 id = 0;
	guint8 *filepath = NULL;
	gint blen = 0;

	if (flags) {};
	if (tree == NULL)
		return;	
	r_tree = proto_item_add_subtree(tree, 
		ett_sar_request);
	/*
	 * Get Transaction ID
	 */
	id = tvb_get_ntohl(tvb, offset);
	proto_tree_add_tranid(r_tree,
		tvb, offset, "Transaction ID",
		id);
	offset += 4;
	/*
	 * Get the file/path name
	 */
	filepath = tvb_get_stringz(tvb, offset, &blen);
	proto_tree_add_string(r_tree, 
		hf_sar_filepath, 
		tvb, offset, blen,
		filepath);
	g_free(filepath);
	offset += blen;
	/*
	 * Get the authentication field if it exists
	 */
	blen = tvb_length_remaining(tvb, offset);
	proto_tree_add_printtext(r_tree, tvb, offset,
		"Authentication: ", blen);
	return;
}

static void
dissect_metadata(guint32 flags, proto_tree *tree, tvbuff_t *tvb, gint offset)
{
	proto_tree *m_tree = NULL;
	guint32 id = 0;
	
	if (tree == NULL)
		return;	
	m_tree = proto_item_add_subtree(tree, 
		ett_sar_metadata);
	/*
	 * Get Transaction ID
	 */
	id = tvb_get_ntohl(tvb, offset);
	proto_tree_add_tranid(m_tree,
		tvb, offset, "Transaction ID",
		id);
	offset += 4;
	/*
	 * Get Checksums
	 */
	switch(SAR_CSUM_TYPE(flags))
	{
	case SAR_CSUM_NONE:
		break;
	case SAR_CSUM_CRC32:
		proto_tree_add_csum(m_tree, tvb, 
			offset, "Checksum CRC32: ", 1);
		offset += 4;
		break;
	case SAR_CSUM_MD5:
		proto_tree_add_csum(m_tree, tvb, 
			offset, "Checksum MD5: ", 4);
		offset += 16;
		break;
	case SAR_CSUM_SHA1:
		proto_tree_add_csum(m_tree, tvb, 
			offset, "Checksum SHA1: ", 5);
		offset += 20;
		break;
	default:
		proto_tree_add_csum(m_tree, tvb, 
			offset, "Checksum UNDEF: ", 
			(gint) SAR_CSUMLEN_TYPE(flags));
		offset += ((gint) SAR_CSUMLEN_TYPE(flags) * 4);
		break;
	}
	offset += proto_tree_add_dir(m_tree, tvb, offset);
	return;
}

static void
dissect_data(guint32 flags, proto_tree *tree, tvbuff_t *tvb, gint offset)
{
	proto_tree *d_tree = NULL;
	guint16 desc_16 = 0;
	guint32 desc_32 = 0;
	guint64 desc_64 = 0;
	guint32 tsec_32 = 0;
	guint64 tsec_64 = 0;
	guint32 id = 0;
	gint blen = 0;
	gint dirlen = 0;
	guint8 *tstring = NULL;
	guint8 *timestamp = NULL;
	struct tm	*t;
	guint8 tb[64];

	if (tree == NULL)
		return;	
	d_tree = proto_item_add_subtree(tree, 
		ett_sar_data);
	/*
	 * Get Transaction ID
	 */
	id = tvb_get_ntohl(tvb, offset);
	proto_tree_add_tranid(d_tree,
		tvb, offset, "Transaction ID",
		id);
	offset += 4;

	/*
	 * Get timestamp info if it is enabled
	 */
	if (SAR_TSTAMP_TYPE(flags) == SAR_TSTAMP_ENABLED)
	{
		timestamp = tvb_get_string(tvb, offset, 16);
		switch(timestamp[15] & SAR_TIMESTAMP_MASK)
		{
		case SAR_TIMESTAMP_32: /* 32 bit utime timestamp */
			tsec_32 = tvb_get_ntohl(tvb, offset);
			t = gmtime((time_t *) &tsec_32);
			asctime_r(t, tb);
			tb[strlen(tb) - 1] = '\0';
			proto_tree_add_text(d_tree, tvb, offset, 
				4, "Timestamp 32 bit: %s", tb);
			break;
		case SAR_TIMESTAMP_64: /* 64 bit utime timestamp */
			tsec_64 = tvb_get_ntoh64(tvb, offset);
			t = gmtime((time_t *) &tsec_64);
			asctime_r(t, tb);
			tb[strlen(tb) - 1] = '\0';
			proto_tree_add_text(d_tree, tvb, offset, 
				4, "Timestamp 64 bit: %s", tb);
			break;
		case SAR_TIMESTAMP_32_32: /* 32 bit utime with 32 bit nanoseconds */
			tsec_32 = tvb_get_ntohl(tvb, offset);
			t = gmtime((time_t *) &tsec_32);
			asctime_r(t, tb);
			tb[strlen(tb) - 1] = '\0';
			tsec_32 = tvb_get_ntohl(tvb, offset + 4);
			proto_tree_add_text(d_tree, tvb, offset, 
				8, "Timestamp 32.32 bit: %s + %-3dms", tb, 
				tsec_32/1000000);
			break;
		case SAR_TIMESTAMP_64_32: /* 64 bit utime with 32 bit nanoseconds */
			tsec_64 = tvb_get_ntoh64(tvb, offset);
			t = gmtime((time_t *) &tsec_64);
			asctime_r(t, tb);
			tb[strlen(tb) - 1] = '\0';
			tsec_32 = tvb_get_ntohl(tvb, offset);
			proto_tree_add_text(d_tree, tvb, offset, 
				12, "Timestamp 64.32 bit: %s + 0.%-3dms", tb, 
				tsec_32/1000000);
			break;
		case SAR_TIMESTAMP_LOCAL: /* Some 128 bit user defined timestamp */
		default:
			tstring = tvb_bytes_to_str_punct(tvb, 
				offset, 16, '-');
			proto_tree_add_text(d_tree, tvb, offset,
				strlen(tstring) + 1,
				"Timestamp 128 bit (Local Interpret): %s\n",
				tstring);
			break;
		}
		offset += 16;
	}

	switch(SAR_DESC_TYPE(flags))
	{
	case SAR_DESC_16:
		desc_16 = tvb_get_ntohs(tvb, offset);
		proto_tree_add_desc16(d_tree,
			tvb, offset, "Offset",
			desc_16);
		offset += 2;
		break;
	case SAR_DESC_32:
		desc_32 = tvb_get_ntohl(tvb, offset);
		proto_tree_add_desc32(d_tree,
			tvb, offset, "Offset",
			desc_32);
		offset += 4;
		break;
	case SAR_DESC_64:
		desc_64 = tvb_get_ntoh64(tvb, offset);
		proto_tree_add_desc64(d_tree,
			tvb, offset, "Offset",
			desc_64);
		offset += 8;
		break;
	case SAR_DESC_128:
		/* No 128 bit support yet!!!! */
		proto_tree_add_desc128(d_tree,
			tvb, offset, "Offset");
		offset += 16;
		break;
	}
	blen = tvb_length_remaining(tvb, offset);
	/*
	 * If we have a _getdir_ or _putdir_ then decode each
	 * of the directory entries in the data payload
	 */
	if (SAR_TFR_TYPE(flags) == SAR_TFR_DIR)
	{
		/*
		 * Go through the directory entries in the DATA frame
		 * Note that directory entries cannot span frames in this code
		 * yet! I need to write some code to handle that!
		 */
		while (blen > 0)
		{
			dirlen = proto_tree_add_dir(d_tree, tvb, offset);
			blen -= dirlen;
			offset += dirlen;
		}
	}
	else
	{
		/*
		 * Rest of Frame is normal DATA transfer of
		 * the Payload just tell us how long it is
		 */
		proto_tree_add_text(d_tree, tvb, offset, blen, 
			"Payload Length: %" G_GINT32_MODIFIER "u",
			blen);
	}
	return;
}

static void
dissect_status(guint32 flags, proto_tree *tree, tvbuff_t *tvb, gint offset)
{
	proto_tree *s_tree = NULL;
	gint i = 0;
	gint blen = 0;
	gint numholes = 0;
	guint32 id = 0;
	guint16 tmp_16 = 0;
	guint16 tmp2_16 = 0;
	guint32 tmp_32 = 0;
	guint32 tmp2_32 = 0;
	guint64 tmp_64 = 0;
	guint64 tmp2_64 = 0;
	guint32 tsec_32 = 0;
	guint64 tsec_64 = 0;
	guint8 *tstring = NULL;
	guint8 *timestamp = NULL;
	struct tm	*t;
	guint8 tb[64];
	
	if (tree == NULL)
		return;	
	s_tree = proto_item_add_subtree(tree, 
		ett_sar_status);

	/*
	 * Get Transaction ID
	 */
	id = tvb_get_ntohl(tvb, offset);
	proto_tree_add_tranid(s_tree,
		tvb, offset, "Transaction ID",
		id);
	offset += 4;
	/*
	 * Get timestamp info if it is enabled
	 */
	if (SAR_TSTAMP_TYPE(flags) == SAR_TSTAMP_ENABLED)
	{
		timestamp = tvb_get_string(tvb, offset, 16);
		switch(timestamp[15] & SAR_TIMESTAMP_MASK)
		{
		case SAR_TIMESTAMP_32: /* 32 bit utime timestamp */
			tsec_32 = tvb_get_ntohl(tvb, offset);
			t = gmtime((time_t *) &tsec_32);
			asctime_r(t, tb);
			tb[strlen(tb) - 1] = '\0';
			proto_tree_add_text(s_tree, tvb, offset, 
				4, "Timestamp 32 bit: %s", tb);
			break;
		case SAR_TIMESTAMP_64: /* 64 bit utime timestamp */
			tsec_64 = tvb_get_ntoh64(tvb, offset);
			t = gmtime((time_t *) &tsec_64);
			asctime_r(t, tb);
			tb[strlen(tb) - 1] = '\0';
			proto_tree_add_text(s_tree, tvb, offset, 
				4, "Timestamp 64 bit: %s", tb);
			break;
		case SAR_TIMESTAMP_32_32: /* 32 bit utime with 32 bit nanoseconds */
			tsec_32 = tvb_get_ntohl(tvb, offset);
			t = gmtime((time_t *) &tsec_32);
			asctime_r(t, tb);
			tb[strlen(tb) - 1] = '\0';
			tsec_32 = tvb_get_ntohl(tvb, offset + 4);
			proto_tree_add_text(s_tree, tvb, offset, 
				8, "Timestamp 32.32 bit: %s + %-3dms", tb, 
				tsec_32/1000000);
			break;
		case SAR_TIMESTAMP_64_32: /* 64 bit utime with 32 bit nanoseconds */
			tsec_64 = tvb_get_ntoh64(tvb, offset);
			t = gmtime((time_t *) &tsec_64);
			asctime_r(t, tb);
			tb[strlen(tb) - 1] = '\0';
			tsec_32 = tvb_get_ntohl(tvb, offset);
			proto_tree_add_text(s_tree, tvb, offset, 
				12, "Timestamp 64.32 bit: %s + 0.%-3dms", tb, 
				tsec_32/1000000);
			break;
		case SAR_TIMESTAMP_LOCAL: /* Some 128 bit user defined timestamp */
		default:
			tstring = tvb_bytes_to_str_punct(tvb, 
				offset, 16, '-');
			proto_tree_add_text(s_tree, tvb, offset,
				strlen(tstring) + 1,
				"Timestamp 128 bit (Local Interpret): %s\n",
				tstring);
			break;
		}
		offset += 16;
	}
	/*
	 * Progress Indicator & In Response To
	 */
	switch(SAR_DESC_TYPE(flags))
	{
	case SAR_DESC_16:
		tmp_16 = tvb_get_ntohs(tvb, offset);
		proto_tree_add_desc16(s_tree,
			tvb, offset, "Current Progress (bytes)",
			tmp_16);
		offset += 2;
		tmp_16 = tvb_get_ntohs(tvb, offset);
		proto_tree_add_desc16(s_tree,
			tvb, offset, "In Response To (bytes)",
			tmp_16);
		offset += 2;
		break;
	case SAR_DESC_32:
		tmp_32 = tvb_get_ntohl(tvb, offset);
		proto_tree_add_desc32(s_tree,
			tvb, offset, "Current Progress (bytes)",
			tmp_32);
		offset += 4;
		tmp_32 = tvb_get_ntohl(tvb, offset);
		proto_tree_add_desc32(s_tree,
			tvb, offset, "In Response To (bytes)",
			tmp_32);
		offset += 4;
		break;
	case SAR_DESC_64:
		tmp_64 = tvb_get_ntoh64(tvb, offset);
		proto_tree_add_desc64(s_tree,
			tvb, offset, "Current Progress (bytes)",
			tmp_64);
		offset += 8;
		tmp_64 = tvb_get_ntoh64(tvb, offset);
		proto_tree_add_desc64(s_tree,
			tvb, offset, "In Response To (bytes)",
			tmp_64);
		offset += 8;
		break;
	case SAR_DESC_128:
		/* No 128 bit support yet!!!! */
		proto_tree_add_desc128(s_tree,
			tvb, offset, "Current Progress (bytes)");
		offset += 16;
		proto_tree_add_desc128(s_tree,
			tvb, offset, "In Response To (bytes)");
		offset += 16;
		break;
	}
	/*
	 * Handle the Holes
	 * Rest of the frame (if any) is holes
	 * Each hole has 2 x descriptors (start and end)
	 */
	blen = tvb_length_remaining(tvb, offset);
	if (blen == 0)
		return;
	switch(SAR_DESC_TYPE(flags))
	{
	case SAR_DESC_16:
		numholes = blen / 4;
		break;
	case SAR_DESC_32:
		numholes = blen / 8;
		break;
	case SAR_DESC_64:
		numholes = blen / 16;
		break;
	case SAR_DESC_128:
		numholes = blen / 32;
		break;
	}
	for (i = 0; i < numholes; i++)
	{
		switch(SAR_DESC_TYPE(flags))
		{
		case SAR_DESC_16:
			tmp_16 = tvb_get_ntohs(tvb, offset);
			tmp2_16 = tvb_get_ntohs(tvb, offset + 2);
			proto_tree_add_hole16(s_tree,
				tvb, offset, i, tmp_16, tmp2_16);
			offset += 4;
			break;
		case SAR_DESC_32:
			tmp_32 = tvb_get_ntohl(tvb, offset);
			tmp2_32 = tvb_get_ntohl(tvb, offset + 4);
			proto_tree_add_hole32(s_tree,
				tvb, offset, i, tmp_32, tmp2_32);
			offset += 8;
			break;
		case SAR_DESC_64:
			tmp_64 = tvb_get_ntoh64(tvb, offset);
			tmp2_64 = tvb_get_ntoh64(tvb, offset + 8);
			proto_tree_add_hole64(s_tree,
				tvb, offset, i, tmp_64, tmp2_64);
			offset += 16;
			break;
		case SAR_DESC_128:
			/* No 128 bit support yet!!!! */
			proto_tree_add_hole128(s_tree,
				tvb, offset, i);
			offset += 32;
			break;
		}
	}
	return;
}

static void
dissect_saratoga(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
{
	proto_item *ti = NULL;
	proto_tree *sar_tree = NULL;
	gint	offset = 0;
	guint	tran = 0;

	guint32 sar_flags = tvb_get_ntohl(tvb, 0);

	guint32	sar_packettype = 
		tvb_get_ntohl(tvb, 0) & SAR_PACKETTYPE_MASK;

	col_set_str(pinfo->cinfo, COL_PROTOCOL, "SARATOGA");
	/* Clear out info column */
	col_clear(pinfo->cinfo, COL_INFO);
	switch(SAR_FRAME_TYPE(sar_flags))
	{
	case SAR_METADATA:
	case SAR_DATA:
	case SAR_STATUS:
		/*
		 * Its useful in the prime window to know what transaction ID
		 * the frame applies to
		 */
		tran = tvb_get_ntohl(tvb, 4);
		col_add_fstr(pinfo->cinfo, COL_INFO, "Flags (%08X) %s ID %08X",
			sar_flags,
			val_to_str(sar_packettype >> SAR_PACKETTYPE_SHIFT, 
				sar_packettype_names, "Type (%01d)"),
			tran);
		break;
	default:
		col_add_fstr(pinfo->cinfo, COL_INFO, "Flags (%08x) %s",
			sar_flags,
			val_to_str(sar_packettype >> SAR_PACKETTYPE_SHIFT, 
				sar_packettype_names, "Type (%01d)"));
		break;
	}
	if (tree == NULL)
		return;
	ti = proto_tree_add_item(tree, proto_saratoga, tvb, 0, -1, FALSE);
	proto_item_append_text(ti, ", Flags: %08x",
			sar_flags);
	proto_item_append_text(ti, ", Type: %s",
		val_to_str(sar_packettype >> SAR_PACKETTYPE_SHIFT, 
			sar_packettype_names, 
			"Unknown (%01d)"));
	switch(SAR_FRAME_TYPE(sar_flags))
	{
	case SAR_METADATA:
	case SAR_DATA:
	case SAR_STATUS:
		proto_item_append_text(ti, ", ID: %08X", tran);
		break;
	default:
		break;
	}
	
	sar_tree = proto_item_add_subtree(ti, ett_saratoga);

	proto_tree_add_item(sar_tree, hf_sar_version, 
		tvb, offset, 4, ENC_BIG_ENDIAN);
	proto_tree_add_item(sar_tree, hf_sar_packettype, 
		tvb, offset, 4, ENC_BIG_ENDIAN);
	proto_tree_add_item(sar_tree, hf_sar_descriptor, 
		tvb, offset, 4, ENC_BIG_ENDIAN);


	/* Handle flags for each of the differing frame types */
	switch(SAR_FRAME_TYPE(sar_flags))
	{
	case SAR_BEACON:
		proto_tree_add_item(sar_tree, hf_sar_bundle,
			tvb, offset, 4, ENC_BIG_ENDIAN);
		proto_tree_add_item(sar_tree, hf_sar_stream,
			tvb, offset, 4, ENC_BIG_ENDIAN);
		proto_tree_add_item(sar_tree, hf_sar_txwill,
			tvb, offset, 4, ENC_BIG_ENDIAN);
		proto_tree_add_item(sar_tree, hf_sar_rxwill,
			tvb, offset, 4, ENC_BIG_ENDIAN);
		proto_tree_add_item(sar_tree, hf_sar_udp,
			tvb, offset, 4, ENC_BIG_ENDIAN);
		proto_tree_add_item(sar_tree, hf_sar_fsp,
			tvb, offset, 4, ENC_BIG_ENDIAN);
		proto_tree_add_item(sar_tree, hf_sar_fspd,
			tvb, offset, 4, ENC_BIG_ENDIAN);
		offset += 4;
		dissect_beacon(sar_flags, sar_tree, tvb, offset);
		break;

	case SAR_REQUEST:
		proto_tree_add_item(sar_tree, hf_sar_bundle,
			tvb, offset, 4, ENC_BIG_ENDIAN);
		proto_tree_add_item(sar_tree, hf_sar_stream,
			tvb, offset, 4, ENC_BIG_ENDIAN);
		proto_tree_add_item(sar_tree, hf_sar_udp,
			tvb, offset, 4, ENC_BIG_ENDIAN);
		proto_tree_add_item(sar_tree, hf_sar_tran,
			tvb, offset, 4, ENC_BIG_ENDIAN);
		offset += 4;
		dissect_request(sar_flags, sar_tree, tvb, offset);
		break;

	case SAR_METADATA:
		proto_tree_add_item(sar_tree, hf_sar_tfr,
			tvb, offset, 4, ENC_BIG_ENDIAN);
		proto_tree_add_item(sar_tree, hf_sar_prog,
			tvb, offset, 4, ENC_BIG_ENDIAN);
		proto_tree_add_item(sar_tree, hf_sar_rel,
			tvb, offset, 4, ENC_BIG_ENDIAN);
		proto_tree_add_item(sar_tree, hf_sar_csumlen,
			tvb, offset, 4, ENC_BIG_ENDIAN);
		proto_tree_add_item(sar_tree, hf_sar_csum,
			tvb, offset, 4, ENC_BIG_ENDIAN);
		offset += 4;
		dissect_metadata(sar_flags, sar_tree, tvb, offset);
		break;

	case SAR_DATA:
		proto_tree_add_item(sar_tree, hf_sar_tfr,
			tvb, offset, 4, ENC_BIG_ENDIAN);
		proto_tree_add_item(sar_tree, hf_sar_tstamp,
			tvb, offset, 4, ENC_BIG_ENDIAN);
		proto_tree_add_item(sar_tree, hf_sar_reqstatus,
			tvb, offset, 4, ENC_BIG_ENDIAN);
		proto_tree_add_item(sar_tree, hf_sar_eod,
			tvb, offset, 4, ENC_BIG_ENDIAN);
		offset += 4;
		dissect_data(sar_flags, sar_tree, tvb, offset);
		break;

	case SAR_STATUS:
		proto_tree_add_item(sar_tree, hf_sar_tstamp,
			tvb, offset, 4, ENC_BIG_ENDIAN);
		proto_tree_add_item(sar_tree, hf_sar_met,
			tvb, offset, 4, ENC_BIG_ENDIAN);
		proto_tree_add_item(sar_tree, hf_sar_allholes,
			tvb, offset, 4, ENC_BIG_ENDIAN);
		proto_tree_add_item(sar_tree, hf_sar_requestby,
			tvb, offset, 4, ENC_BIG_ENDIAN);
		proto_tree_add_item(sar_tree, hf_sar_statusflag,
			tvb, offset, 4, ENC_BIG_ENDIAN);
		offset += 4;
		dissect_status(sar_flags, sar_tree, tvb, offset);
		break;

	default:
		break;
	}
}

void
proto_reg_handoff_saratoga(void)
{
	static dissector_handle_t	sar_handle;

	sar_handle = create_dissector_handle(dissect_saratoga, 
		proto_saratoga);
	dissector_add_uint("udp.port", SAR_PORT, sar_handle);
}

/*
 * $Log: packet-saratoga.c,v $
 * Revision 1.4  2011/08/05 05:36:58  charlesetsmith
 * FIxed bug outputting formatted text strings. It basically'
 * wasn;t outputting them. SO if the chracters are printable
 * it outputs them. If not a series of hex strings.
 *
 * Revision 1.2  2011/07/25 02:15:47  charlesetsmith
 * Sorted out ctime and mtime entries to be compliant with the unsigned
 * 32 bit integer value of number of seconds since 1/1/2000 as used in the
 * bundle protocol.
 * Display of transaction id is 8 hex digits.
 * String output for EID to print the string if it is
 * a printable char set or hex if not.
 *
 * Revision 1.1.1.1  2011/07/23 02:27:00  charlesetsmith
 * Saratoga Wireshark packet dissector
 *
 * Revision 1.1  2011/07/23 01:04:54  charlesetsmith
 * Saratoga wireshark dissector
 *
 *
 */

